@@ -24,6 +24,7 @@
 #include <unistd.h>
 #include <dirent.h>
 #include <uuid/uuid.h>
+#include <getopt.h>
 
 #include "kerncompat.h"
 #include "ctree.h"
@@ -40,6 +41,7 @@ static int update_seeding_flag(struct btrfs_root *root, int set_flag)
 	struct btrfs_trans_handle *trans;
 	struct btrfs_super_block *disk_super;
 	u64 super_flags;
+	int ret;
 
 	disk_super = root->fs_info->super_copy;
 	super_flags = btrfs_super_flags(disk_super);
@@ -64,41 +66,26 @@ static int update_seeding_flag(struct btrfs_root *root, int set_flag)
 
 	trans = btrfs_start_transaction(root, 1);
 	btrfs_set_super_flags(disk_super, super_flags);
-	btrfs_commit_transaction(trans, root);
+	ret = btrfs_commit_transaction(trans, root);
 
-	return 0;
+	return ret;
 }
 
-static int enable_extrefs_flag(struct btrfs_root *root)
+static int set_super_incompat_flags(struct btrfs_root *root, u64 flags)
 {
 	struct btrfs_trans_handle *trans;
 	struct btrfs_super_block *disk_super;
 	u64 super_flags;
+	int ret;
 
 	disk_super = root->fs_info->super_copy;
 	super_flags = btrfs_super_incompat_flags(disk_super);
-	super_flags |= BTRFS_FEATURE_INCOMPAT_EXTENDED_IREF;
+	super_flags |= flags;
 	trans = btrfs_start_transaction(root, 1);
 	btrfs_set_super_incompat_flags(disk_super, super_flags);
-	btrfs_commit_transaction(trans, root);
+	ret = btrfs_commit_transaction(trans, root);
 
-	return 0;
-}
-
-static int enable_skinny_metadata(struct btrfs_root *root)
-{
-	struct btrfs_trans_handle *trans;
-	struct btrfs_super_block *disk_super;
-	u64 super_flags;
-
-	disk_super = root->fs_info->super_copy;
-	super_flags = btrfs_super_incompat_flags(disk_super);
-	super_flags |= BTRFS_FEATURE_INCOMPAT_SKINNY_METADATA;
-	trans = btrfs_start_transaction(root, 1);
-	btrfs_set_super_incompat_flags(disk_super, super_flags);
-	btrfs_commit_transaction(trans, root);
-
-	return 0;
+	return ret;
 }
 
 static int change_header_uuid(struct btrfs_root *root, struct extent_buffer *eb)
@@ -267,7 +254,9 @@ out:
 
 static int change_fsid_prepare(struct btrfs_fs_info *fs_info)
 {
+	struct btrfs_root *tree_root = fs_info->tree_root;
 	u64 flags = btrfs_super_flags(fs_info->super_copy);
+	int ret = 0;
 
 	if (!fs_info->new_fsid && !fs_info->new_chunk_tree_uuid)
 		return 0;
@@ -276,7 +265,16 @@ static int change_fsid_prepare(struct btrfs_fs_info *fs_info)
 		flags |= BTRFS_SUPER_FLAG_CHANGING_FSID;
 	btrfs_set_super_flags(fs_info->super_copy, flags);
 
-	return write_all_supers(fs_info->tree_root);
+	memcpy(fs_info->super_copy->fsid, fs_info->new_fsid, BTRFS_FSID_SIZE);
+	ret = write_all_supers(tree_root);
+	if (ret < 0)
+		return ret;
+
+	/* also restore new chunk_tree_id into tree_root for restore */
+	write_extent_buffer(tree_root->node, fs_info->new_chunk_tree_uuid,
+			    btrfs_header_chunk_tree_uuid(tree_root->node),
+			    BTRFS_UUID_SIZE);
+	return write_tree_block(NULL, tree_root, tree_root->node);
 }
 
 static int change_fsid_done(struct btrfs_fs_info *fs_info)
@@ -293,42 +291,78 @@ static int change_fsid_done(struct btrfs_fs_info *fs_info)
 	return write_all_supers(fs_info->tree_root);
 }
 
-static int change_uuid(struct btrfs_fs_info *fs_info, const char *new_fsid,
-		       const char *new_chunk_uuid)
+/*
+ * Return 0 for no unfinished fsid change.
+ * Return >0 for unfinished fsid change, and restore unfinished fsid/
+ * chunk_tree_id into fsid_ret/chunk_id_ret.
+ */
+static int check_unfinished_fsid_change(struct btrfs_fs_info *fs_info,
+					uuid_t fsid_ret, uuid_t chunk_id_ret)
 {
+	struct btrfs_root *tree_root = fs_info->tree_root;
+	u64 flags = btrfs_super_flags(fs_info->super_copy);
+
+	if (flags & BTRFS_SUPER_FLAG_CHANGING_FSID) {
+		memcpy(fsid_ret, fs_info->super_copy->fsid, BTRFS_FSID_SIZE);
+		read_extent_buffer(tree_root->node, chunk_id_ret,
+				btrfs_header_chunk_tree_uuid(tree_root->node),
+				BTRFS_UUID_SIZE);
+		return 1;
+	}
+	return 0;
+}
+
+/*
+ * Change fsid of a given fs.
+ *
+ * If new_fsid_str is not given, use a random generated UUID.
+ * Caller should check new_fsid_str is valid
+ */
+static int change_uuid(struct btrfs_fs_info *fs_info, const char *new_fsid_str)
+{
+	uuid_t new_fsid;
+	uuid_t new_chunk_id;
+	uuid_t old_fsid;
+	char uuid_buf[BTRFS_UUID_UNPARSED_SIZE];
 	int ret = 0;
 
-	/* caller should do extra check on passed uuid */
-	if (new_fsid) {
-		/* allocated mem will be freed at close_ctree() */
-		fs_info->new_fsid = malloc(BTRFS_FSID_SIZE);
-		if (!fs_info->new_fsid) {
-			ret = -ENOMEM;
-			goto out;
-		}
-		ret = uuid_parse(new_fsid, fs_info->new_fsid);
-		if (ret < 0)
-			goto out;
-	}
+	if (check_unfinished_fsid_change(fs_info, new_fsid, new_chunk_id)) {
+		if (new_fsid_str) {
+			uuid_t tmp;
 
-	if (new_chunk_uuid) {
-		/* allocated mem will be freed at close_ctree() */
-		fs_info->new_chunk_tree_uuid = malloc(BTRFS_UUID_SIZE);
-		if (!fs_info->new_chunk_tree_uuid) {
-			ret = -ENOMEM;
-			goto out;
+			uuid_parse(new_fsid_str, tmp);
+			if (memcmp(tmp, new_fsid, BTRFS_FSID_SIZE)) {
+				fprintf(stderr,
+		"ERROR: New fsid %s is not the same with unfinished fsid change\n",
+					new_fsid_str);
+				return -EINVAL;
+			}
 		}
-		ret = uuid_parse(new_chunk_uuid, fs_info->new_chunk_tree_uuid);
-		if (ret < 0)
-			goto out;
-	}
+	} else {
+		if (new_fsid_str)
+			uuid_parse(new_fsid_str, new_fsid);
+		else
+			uuid_generate(new_fsid);
 
+		uuid_generate(new_chunk_id);
+	}
+	fs_info->new_fsid = new_fsid;
+	fs_info->new_chunk_tree_uuid = new_chunk_id;
+
+	memcpy(old_fsid, (const char*)fs_info->fsid, BTRFS_UUID_SIZE);
+	uuid_unparse(old_fsid, uuid_buf);
+	printf("Current fsid: %s\n", uuid_buf);
+
+	uuid_unparse(new_fsid, uuid_buf);
+	printf("New fsid: %s\n", uuid_buf);
 	/* Now we can begin fsid change */
+	printf("Set superblock flag CHANGING_FSID\n");
 	ret = change_fsid_prepare(fs_info);
 	if (ret < 0)
 		goto out;
 
 	/* Change extents first */
+	printf("Change fsid in extents\n");
 	ret = change_extents_uuid(fs_info);
 	if (ret < 0) {
 		fprintf(stderr, "Failed to change UUID of metadata\n");
@@ -336,25 +370,28 @@ static int change_uuid(struct btrfs_fs_info *fs_info, const char *new_fsid,
 	}
 
 	/* Then devices */
+	printf("Change fsid on devices\n");
 	ret = change_devices_uuid(fs_info);
 	if (ret < 0) {
 		fprintf(stderr, "Failed to change UUID of devices\n");
 		goto out;
 	}
 
-	/* Last, change fsid in super, only fsid change needs this */
-	if (new_fsid) {
-		memcpy(fs_info->fs_devices->fsid, fs_info->new_fsid,
-		       BTRFS_FSID_SIZE);
-		memcpy(fs_info->super_copy->fsid, fs_info->new_fsid,
-		       BTRFS_FSID_SIZE);
-		ret = write_all_supers(fs_info->tree_root);
-		if (ret < 0)
-			goto out;
-	}
+	/* Last, change fsid in super */
+	memcpy(fs_info->fs_devices->fsid, fs_info->new_fsid,
+	       BTRFS_FSID_SIZE);
+	memcpy(fs_info->super_copy->fsid, fs_info->new_fsid,
+	       BTRFS_FSID_SIZE);
+	ret = write_all_supers(fs_info->tree_root);
+	if (ret < 0)
+		goto out;
 
 	/* Now fsid change is done */
+	printf("Clear superblock flag CHANGING_FSID\n");
 	ret = change_fsid_done(fs_info);
+	fs_info->new_fsid = NULL;
+	fs_info->new_chunk_tree_uuid = NULL;
+	printf("Fsid change finished\n");
 out:
 	return ret;
 }
@@ -365,23 +402,33 @@ static void print_usage(void)
 	fprintf(stderr, "\t-S value\tpositive value will enable seeding, zero to disable, negative is not allowed\n");
 	fprintf(stderr, "\t-r \t\tenable extended inode refs\n");
 	fprintf(stderr, "\t-x \t\tenable skinny metadata extent refs\n");
-	fprintf(stderr, "\t-f \t\tforce to set or clear flags, make sure that you are aware of the dangers\n");
+	fprintf(stderr, "\t-n \t\tenable no-holes feature (more efficient sparse file representation)\n");
+	fprintf(stderr, "\t-f \t\tforce to do dangerous operation, make sure that you are aware of the dangers\n");
+	fprintf(stderr, "\t-u \t\tchange fsid, use a random one\n");
+	fprintf(stderr, "\t-U UUID\t\tchange fsid to UUID\n");
 }
 
 int main(int argc, char *argv[])
 {
 	struct btrfs_root *root;
+	enum btrfs_open_ctree_flags ctree_flags = OPEN_CTREE_WRITES;
 	int success = 0;
 	int total = 0;
-	int extrefs_flag = 0;
 	int seeding_flag = 0;
 	u64 seeding_value = 0;
-	int skinny_flag = 0;
+	int random_fsid = 0;
+	char *new_fsid_str = NULL;
 	int ret;
+	u64 super_flags = 0;
 
 	optind = 1;
 	while(1) {
-		int c = getopt(argc, argv, "S:rxf");
+		static const struct option long_options[] = {
+			{ "help", no_argument, NULL, GETOPT_VAL_HELP},
+			{ NULL, 0, NULL, 0 }
+		};
+		int c = getopt_long(argc, argv, "S:rxfuU:n", long_options, NULL);
+
 		if (c < 0)
 			break;
 		switch(c) {
@@ -390,17 +437,29 @@ int main(int argc, char *argv[])
 			seeding_value = arg_strtou64(optarg);
 			break;
 		case 'r':
-			extrefs_flag = 1;
+			super_flags |= BTRFS_FEATURE_INCOMPAT_EXTENDED_IREF;
 			break;
 		case 'x':
-			skinny_flag = 1;
+			super_flags |= BTRFS_FEATURE_INCOMPAT_SKINNY_METADATA;
+			break;
+		case 'n':
+			super_flags |= BTRFS_FEATURE_INCOMPAT_NO_HOLES;
 			break;
 		case 'f':
 			force = 1;
 			break;
+		case 'U':
+			ctree_flags |= OPEN_CTREE_IGNORE_FSID_MISMATCH;
+			new_fsid_str = optarg;
+			break;
+		case 'u':
+			ctree_flags |= OPEN_CTREE_IGNORE_FSID_MISMATCH;
+			random_fsid = 1;
+			break;
+		case GETOPT_VAL_HELP:
 		default:
 			print_usage();
-			return 1;
+			return c != GETOPT_VAL_HELP;
 		}
 	}
 
@@ -412,13 +471,36 @@ int main(int argc, char *argv[])
 		return 1;
 	}
 
-	if (!(seeding_flag + extrefs_flag + skinny_flag)) {
+	if (random_fsid && new_fsid_str) {
+		fprintf(stderr,
+			"ERROR: Random fsid can't be used with specified fsid\n");
+		return 1;
+	}
+	if (!super_flags && !seeding_flag && !(random_fsid || new_fsid_str)) {
 		fprintf(stderr,
 			"ERROR: At least one option should be assigned.\n");
 		print_usage();
 		return 1;
 	}
 
+	if (new_fsid_str) {
+		uuid_t tmp;
+
+		ret = uuid_parse(new_fsid_str, tmp);
+		if (ret < 0) {
+			fprintf(stderr,
+				"ERROR: Could not parse UUID: %s\n",
+				new_fsid_str);
+			return 1;
+		}
+		if (!test_uuid_unique(new_fsid_str)) {
+			fprintf(stderr,
+				"ERROR: Fsid %s is not unique\n",
+				new_fsid_str);
+			return 1;
+		}
+	}
+
 	ret = check_mounted(device);
 	if (ret < 0) {
 		fprintf(stderr, "Could not check mount status: %s\n",
@@ -429,7 +511,7 @@ int main(int argc, char *argv[])
 		return 1;
 	}
 
-	root = open_ctree(device, 0, OPEN_CTREE_WRITES);
+	root = open_ctree(device, 0, ctree_flags);
 
 	if (!root) {
 		fprintf(stderr, "Open ctree failed\n");
@@ -442,7 +524,8 @@ int main(int argc, char *argv[])
 			ret = ask_user("We are going to clear the seeding flag, are you sure?");
 			if (!ret) {
 				fprintf(stderr, "Clear seeding flag canceled\n");
-				return 1;
+				ret = 1;
+				goto out;
 			}
 		}
 
@@ -452,15 +535,29 @@ int main(int argc, char *argv[])
 		total++;
 	}
 
-	if (extrefs_flag) {
-		enable_extrefs_flag(root);
-		success++;
+	if (super_flags) {
+		ret = set_super_incompat_flags(root, super_flags);
+		if (!ret)
+			success++;
 		total++;
 	}
 
-	if (skinny_flag) {
-		enable_skinny_metadata(root);
-		success++;
+	if (random_fsid || new_fsid_str) {
+		if (!force) {
+			fprintf(stderr,
+				"Warning: It's highly recommended to run 'btrfs check' before this operation\n");
+			fprintf(stderr,
+				"Also canceling running UUID change progress may cause corruption\n");
+			ret = ask_user("We are going to change UUID, are your sure?");
+			if (!ret) {
+				fprintf(stderr, "UUID change canceled\n");
+				ret = 1;
+				goto out;
+			}
+		}
+		ret = change_uuid(root->fs_info, new_fsid_str);
+		if (!ret)
+			success++;
 		total++;
 	}
 
@@ -471,6 +568,7 @@ int main(int argc, char *argv[])
 		ret = 1;
 		fprintf(stderr, "btrfstune failed\n");
 	}
+out:
 	close_ctree(root);
 
 	return ret;
